/*

PILOTAGE D'UNE PLATE FORME 6 AXES DE TYPE 2 

1 - Dans le menu Outils/Type de carte choisir Arduino Mega 
2 - Dans le menu Outil/Port choisir le port sur lequel est branché l'Arduino Mega
3 - Téléverser ce programme sur l'Arduino Mega 
4a - Démarrer le moniteur série 
   - Ecrire les commandes suivantes puis valider à chaque fois (le point virgule est important) : 
     carre;
     a=40;
     n=4;
     z=180;
     Kp=20;
     Ki=0.05;
     Kd=34;
     tempo=2000;
     tol=1;
     Start;
   - La plateforme doit effectuer une trajectoire carrée 
4b - Démarrer le logiciel fourni pour piloter la plateforme 
   - Choisir le port série, se connecter
   - Choisir puis définir une trajectoire, un nombre de points
   - Définir les paramètres de l'asservissement (laisser les paramètres par défaut pour commencer) 
   - Exécuter le mouvement 
   - Tracer les courbes 
   
F. LOUF     --- ENS PARIS SACLAY ---
L. RICHARD  --- ENS PARIS SACLAY ---

Mars 2018 
  
*/

#define beta   5.453*PI/180.
#define Rf     166.0115
#define Rm     152.4914
#define Lb     200.0 
#define Lmin   210.0
#define Xmin   20.0
#define Xmax   135.0
#define L0     191.1668
#define nmax   100

// Saturation de l'intégrale
#define INTG_MAX  10

// PINs correspondant aux ports I1 I2 de chaque driver moteur
int pinI1I2[12] = {22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33} ;

// PINs correspondant aux ports PWM de chaque driver moteur
int speedpin[6] = {2, 3, 4, 5, 6, 7} ; 

// PINs correspondant aux mesures de position (potentiomètres)
int pinpot[6] = {A0, A1, A2, A3, A4, A5} ;

// Variables globales
int pospot ;
float Te = 5;
int com ;

// Variable pour la communication 
String  Ser_Inp_Str = "";          // Cette chaine représente ce qui est arrivé et à decoder
boolean Ser_Inp_Str_End  = false;  // Booleen qui indique la chaine précédente est complète et à décoder   
char end_of_com = ';' ;            // Délimiteur de commande arrivée sur le port série (compatible avec interface)

// Variables associées à l'asservissement et valeurs par défaut
float Kp = 0.0 ; 
float Ki = 0.0 ;
float Kd = 0.0 ;  
float TOL_ERR = 1.5;
float Tempo = 4000;

// Paramètres liées à la régulation (par défaut)
float KP[6] = {Kp,Kp,Kp,Kp,Kp,Kp} ;
float KD[6] = {Ki,Ki,Ki,Ki,Ki,Ki} ;
float KI[6] = {Kd,Kd,Kd,Kd,Kd,Kd} ;
float error[6]  = {0,0,0,0,0,0}; 
float error0[6] = {0,0,0,0,0,0}; 
float integr[6] = {0,0,0,0,0,0}; 
float deriv[6]  = {0,0,0,0,0,0};

// Variables associées aux trajectoires et valeurs par défaut
float a0 = 0.0 ;
float r0 = 0.0 ;
float R0 = 0.0 ;
float p0 = 0.0 ;
float z0 = 180.0 ;
float Z0 = 180.0 ;
int n = 1 ;
int traj_type = 1 ; 

// Variables pour le démarrage/arrêt
boolean start = false ; 

// Points de la trajectoire (position et orientation)
float tabposx[nmax] ;
float tabposy[nmax] ;
float tabposz[nmax] ;
float tabposth[nmax] ;
float tabposph[nmax] ;
float tabposps[nmax] ;


// =========================== INITIALISATION =================================================

void setup() {

  // Boucle sur le moteurs, type des pin associées
  for (int im=0; im<6; im++){
    pinMode(pinI1I2[2*im  ],OUTPUT);
    pinMode(pinI1I2[2*im+1],OUTPUT);   
    pinMode(speedpin[ im  ],OUTPUT);
    }
    
  // Ouverture du port série 
  Serial.begin(115200);
  while (!Serial) {
    ; // Attente de connection au port série (pas forcément utile)
  }
  delay(1000) ;

  // On lit tout ce qui arrive sur le port série et on décode si besoin
  while(!start) {
    serial_input_read() ; 
    if (Ser_Inp_Str_End) {
      serial_input_decode();
    } 
  }
  
  Serial.println("* On peut demarrer !! ");
  
}
// ============================================================================================


// =========================== LECTURE DE CHAINE ARRIVANT SUR LE PORT SERIE ===================

void serial_input_read() {
  while (Serial.available()) {
    // On recupere le caractere
    char inChar = (char)Serial.read();
    // Si on a recupere une fin de commande par le caractere end_of_com, 
    // on met Ser_Inp_Str_End a vrai; 
    if (inChar == end_of_com) {
      Ser_Inp_Str_End = true; 
    }
    else {
      if (inChar != '\n' ){
          // Ajout du caractere a la chaine
          Ser_Inp_Str += inChar;
      }
    }
  }
}
// ============================================================================================

// =========================== CONSTRUCTION DES POINTS DE LA TRAJECTOIRE =====================

void build_traj(){

  float u ;
       
  switch (traj_type){

    // Trajectoire circulaire 
    case 1:
      for (int ip=0; ip<n; ip++)
      {
        tabposx[ip]  = r0*cos(ip*2*PI/n) ;
        tabposy[ip]  = r0*sin(ip*2*PI/n) ;
        tabposz[ip]  = z0 ;
        tabposph[ip] = 0. ;
        tabposth[ip] = 0. ;
        tabposps[ip] = 0. ; 
       } 
       break;

    // Trajectoire carrée   
    case 2:
      n = 4*floor(n/4.0) ;
      for (int ip=0; ip<(n/4); ip++)
      {
        tabposx[ip]  = (a0/2.0) ;
        tabposy[ip]  = (a0/2.0)*(-1 + 8*ip/float(n)) ;
        tabposx[ip+n/4]  = (a0/2.0)*(1 - 8*ip/float(n)) ;
        tabposy[ip+n/4]  = (a0/2.0) ;
        tabposx[ip+n/2]  = -(a0/2.0) ;
        tabposy[ip+n/2]  =  (a0/2.0)*(1 - 8*ip/float(n)) ;        
        tabposx[ip+3*n/4] =  (a0/2.0)*(-1 + 8*ip/float(n)) ;
        tabposy[ip+3*n/4] = -(a0/2.0) ;
      }
      for (int ip=0; ip<n; ip++){
        tabposz[ip]  = z0 ;
        tabposph[ip] = 0. ;
        tabposth[ip] = 0. ;
        tabposps[ip] = 0. ; 
       }   
       break;

    // Trajectoire spirale 
    case 3:
      for (int ip=0; ip<n; ip++)
      {
        u = ip*(R0-r0)/p0/(n-1)  ;
        tabposx[ip]  = (r0+p0*u)*cos(2*PI*u) ;
        tabposy[ip]  = (r0+p0*u)*sin(2*PI*u) ;
        tabposz[ip]  = z0 ;
        tabposph[ip] = 0. ;
        tabposth[ip] = 0. ;
        tabposps[ip] = 0. ; 
       } 
       break;

    // Trajectoire helicoidale 
    case 4:
      for (int ip=0; ip<n; ip++)
      {
        u = ip*(Z0-z0)/p0/(n-1)  ;
        tabposx[ip]  = r0*cos(2*PI*u) ;
        tabposy[ip]  = r0*sin(2*PI*u) ;
        tabposz[ip]  = z0 + p0*u ;
        tabposph[ip] = 0. ;
        tabposth[ip] = 0. ;
        tabposps[ip] = 0. ; 
       } 
       break;
    }
}
// ============================================================================================

// =========================== DECODAGE DES MOTS CLE SUR LE PORT SERIE ========================

void serial_input_decode(){
  
  boolean com_known = false ; 
  int ind_equal ;
  String str_val ; 
  float val ; 

  // On enleve le retour chariot à la fin 
  Ser_Inp_Str = Ser_Inp_Str.substring(0,Ser_Inp_Str.length()) ; 
  
  // On enleve d'eventuels espaces 
  Ser_Inp_Str.replace(" ","") ;
  
  // On cherche un éventuel signe = dans la commande 
  ind_equal = Ser_Inp_Str.indexOf("=",0) ; 
  if (ind_equal >= 0) {
    // On separe alors le mot clé de la valeur associee
    str_val = Ser_Inp_Str.substring(ind_equal+1,Ser_Inp_Str.length()) ; 
    val = str_val.toFloat();  
    Ser_Inp_Str = Ser_Inp_Str.substring(0,ind_equal) ; 
  }

  // Lecture d'un start pour démarrer le pilotage  
  if (Ser_Inp_Str == "Start") {
     Serial.println( "* Demarrage de l'asservissement") ; 
     start = true ; 
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  } 

  // Lecture d'un stop pour cesser le pilotage  (inutilisé actuellement)
  if (Ser_Inp_Str ==  "Stop") {
     Serial.println( "* Arrêt de l'asservissement") ;
     start = false ; 
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  }
  
// Lecture d'un type de trajectoire : cercle
  if (Ser_Inp_Str ==  "cercle") {
     Serial.println( "* Trajectoire plane circulaire") ;
     traj_type = 1 ;  
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  }
  
// Lecture d'un type de trajectoire : carre
  if (Ser_Inp_Str ==  "carre") {
     Serial.println( "* Trajectoire plane carrée") ;
     traj_type = 2 ;  
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  }

// Lecture d'un type de trajectoire : spirale
  if (Ser_Inp_Str ==  "spirale") {
     Serial.println( "* Trajectoire plane en spirale") ;
     traj_type = 3 ;  
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  }

// Lecture d'un type de trajectoire : hélicoide
  if (Ser_Inp_Str ==  "helicoide") {
     Serial.println( "* Trajectoire plane helicoidale") ;
     traj_type = 4 ;  
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  }

// Lecture du temps a partir duquel on passe au point suivant 
  if (Ser_Inp_Str ==  "tempo") {
     Tempo = val ;
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  }

// Lecture du nombre de points de la trajectoire  
  if (Ser_Inp_Str ==  "n") {
     n = int(val) ; 
     if (n>nmax) {
        n=nmax;
     }
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  } 

// Lecture d'une altitude z (altitude des trajectoires planes ou fin hélicoide)
  if (Ser_Inp_Str ==  "z") {
     z0 = val ; 
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  }   

// Lecture d'une altitude Z (altitude fin helicoide)
  if (Ser_Inp_Str ==  "Z") {
     Z0 = val ; 
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  }   

// Lecture d'un coté a (cote d'un carre)
  if (Ser_Inp_Str ==  "a") {
     a0 = val ; 
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  } 

// Lecture d'un rayon r (rayon cercle, rayon début spirale, rayon hélicoide)
  if (Ser_Inp_Str ==  "r") {
     r0 = val ; 
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  } 

// Lecture d'un rayon R (rayon fin spirale)
  if (Ser_Inp_Str ==  "R") {
     R0 = val ; 
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  } 

// Lecture d'un pas (spirale ou hélicoide) 
  if (Ser_Inp_Str ==  "p") {
     p0 = val ; 
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  } 

// Lecture de la tolérance servant de critère d'arrêt  
  if (Ser_Inp_Str ==  "tol") {
     TOL_ERR = val ; 
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  }

// Lecture d'un coefficient Kp  
  if (Ser_Inp_Str ==  "Kp") {
     Kp = val ; 
     for (int im = 0; im<6; im++){
      KP[im] = Kp;
     }
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  }

// Lecture d'un coefficient Kd  
    if (Ser_Inp_Str ==  "Kd") {
     Kd = val ; 
     for (int im = 0; im<6; im++){
      KD[im] = Kd;
     }
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  }

// Lecture d'un coefficient Ki    
  if (Ser_Inp_Str ==  "Ki") {
     Ki = val ; 
     for (int im = 0; im<6; im++){
      KI[im] = Ki;
     }
     Ser_Inp_Str = "";
     Ser_Inp_Str_End = false ;
     com_known = true ; 
  }
  
// Cas d'un mot clé inconnu  
  if (!com_known) {
      Ser_Inp_Str_End = false ;
      Ser_Inp_Str = "";
  }  

 }
// ============================================================================================

// =========================== CALCUL ET VERIFICATION DES LONGUEURS VERINS ====================

// Calcule les 2 racines réelles si elles existent d'une équation du second degré
// Vérifie que les solutions sont possibles 
// Arguments : - a,b,c : paramètres de l'équation a x^2 + bx + c = 0
//             - i     : numéro du vérin
//             - Lv[6] : longueur de chaque vérin 
// Sorties   : - Lv[6] : longueur de chaque vérin modifié pour le verin i

void IOcheck(float a, float b, float c, int i, float Lv[6]){

float delta ;
float x ; 
float y ; 

delta = pow(b,2) - 4.0*a*c ; 

if (delta > 0) { 
  x = (-b+sqrt(delta))/(2*a) ; 
  y = (-b-sqrt(delta))/(2*a) ; 
  if ((x>=Xmin ) and (x <= Xmax)) {   
    Lv[i] = x ;
  } 
  else if ((y>=Xmin ) and (y <= Xmax)) {   
    Lv[i] = y ;
  } 
  else if  (x > Xmax) {
    Lv[i] = Xmax;
    }
  else if  (x < Xmin) {
    Lv[i] = Xmin ;
  }
  else { 
    Lv[i] = (Xmin+Xmax)/2.0 ; 
  }
} 
else
{ 
  Lv[i] = Xmin  ;  
} 
  
}
// ============================================================================================

// =========================== LOI ENTREE SORTIE INVERSE ====================================== 
// Calcule les longueurs vérins pour une configuration (x y z phi theta psi) 
// Arguments : - position plateforme (x, y, z) in mm et 
//               orientation plateforme (phi, theta, psi) in deg
// Sorties   : - longueurs des vérins (x6) in mm

void IOLaw(float X, float Y, float Z, float theta, float phi, float psi, float Lv[6]){

// Local variables
  float A ;
  float B ;
  float C ;
  float a ; 
  float b ; 
  float c ; 

//  Serial.println("* -----------------------------------------------") ; 
//  Serial.print("* Desired position Om (x, y, z ) : ") ; 
//  Serial.print(X) ;
//  Serial.print(" ") ;
//  Serial.print(Y) ;
//  Serial.print(", ") ;
//  Serial.println(Z) ; 
//  Serial.print("* Desired orientation (Th,Ph,Ps) : ") ; 
//  Serial.print(theta) ;
//  Serial.print(", ") ;
//  Serial.print(phi) ;
//  Serial.print(", ") ;
//  Serial.println(psi) ;

// Transformation des deg en rad
   theta = theta*PI/180.0 ;
   phi   =   phi*PI/180.0 ;
   psi   =   psi*PI/180.0 ;  

// On écrit les équations vérin par vérin 
// Vérin 1
   A = Y +  Rm*cos(beta -psi)*sin(phi) - Rm*cos(theta)*cos(phi)*sin(beta -psi) ;
   B = 0. ; 
   C = pow(Lb,2) - pow(Z - Rm*sin(theta)*sin(beta-psi),2) - pow(-Rf + X + Rm*cos(phi)*cos(beta -psi) + Rm*cos(theta)*sin(phi)*sin(beta-psi),2)  ;
   a = 1.0 ; 
   b = 2.0*A ; 
   c = pow(A,2) - C ; 
   IOcheck(a, b, c, 0, Lv) ; 
 
// Vérin 2
   A = Y + Rm*cos(beta+psi)*sin(phi) + Rm*cos(theta)*cos(phi)*sin(beta+psi) ; 
   B = 0.  ;
   C = pow(Lb,2) - pow(Z + Rm*sin(theta)*sin(beta+psi),2) - pow(Rf - X - Rm*cos(phi)*cos(beta+psi) + Rm*cos(theta)*sin(phi)*sin(beta+psi),2) ;
   a = 1.0 ; 
   b = -2.0*A ; 
   c = pow(A,2) - C ;
   IOcheck(a, b, c, 1, Lv) ; 

// Vérin 3
   A = sqrt(3.0)*Rf - 2*Y - 2*Rm*cos(theta)*cos(phi)*cos(PI/6.0-beta+psi) + Rm*sin(phi)*(2*cos(psi)*sin(PI/6.0 -beta)+(sqrt(3.0)*cos(beta)+sin(beta))*sin(psi)) ;
   B = Rf + 2*X  - Rm*(2*cos(theta)*cos(PI/6.0-beta+psi)*sin(phi)+sin(PI/6.0-beta-phi+psi)+sin(PI/6.0-beta+phi+psi)) ;
   C = 4.0*pow(Lb,2) - 4.0*pow(Z + Rm*cos(PI/6.0-beta+psi)*sin(theta),2) ; 
   a = 4.0 ;
   b = 2.0*A - 2*sqrt(3.0)*B ;
   c = pow(A,2) + pow(B,2) - C ;
   IOcheck(a, b, c, 2, Lv) ; 

// Vérin 4
   A = -sqrt(3.0)*Rf + 2*Y - 2*Rm*sin(PI/6.0+beta)*(cos(psi)*sin(phi) + cos(theta)*cos(phi)*sin(psi))+2*Rm*cos(PI/6.0+beta)*(cos(theta)*cos(phi)*cos(psi)-sin(phi)*sin(psi)) ; 
   B = Rf + 2*X- Rm*(2*cos(theta)*cos(PI/6.0+beta+psi)*sin(phi)+sin(PI/6.0+beta -phi+psi) + sin(PI/6.0+beta+phi+psi )) ; 
   C = 4.0*pow(Lb,2) - 4.0*pow(Z + Rm*cos(PI/6.0+beta+psi)*sin(theta),2) ;
   a = 4.0 ;
   b = 2.0*A + 2.0*sqrt(3.0)*B ;
   c = pow(A,2) + pow(B,2) - C ;
   IOcheck(a, b, c, 3, Lv) ; 

// Vérin 5
   A = Rf + 2.0*X + 2.0*Rm*cos(theta)*cos(PI/6.0+beta-psi)*sin(phi) - 2.0*Rm*cos(phi) *sin(PI/6.0 +beta - psi) ; 
   B = 2.0*Rm*cos(PI/6.0+beta)*(cos(theta)*cos(phi)*cos(psi) - sin(phi)*sin(psi)) - sqrt(3.0)*Rf - 2.0*Y+ 2.0*Rm*sin(PI/6.0+beta)*(cos(psi)*sin(phi) + cos(theta)*cos(phi)*sin(psi)) ;
   C = 4.0*pow(Lb,2) - 4.0*pow(Z - Rm*cos(PI/6.0+beta-psi)*sin(theta),2) ; 
   a = 4.0 ;
   b = 2.0*B + 2.0*sqrt(3.0)*A ;
   c = pow(A,2) + pow(B,2) - C ;
   IOcheck(a, b, c, 4, Lv) ; 

// Vérin 6
   A = sqrt(3.0)*Rf + 2.0*Y - 2.0*Rm*sin(PI/6.0-beta)*(cos(psi)*sin(phi) + cos(theta)*cos(phi)*sin(psi))+ 2.0*Rm*cos(PI/6.0-beta)*(-cos(theta)*cos(phi)*cos(psi) + sin(phi)*sin(psi)) ; 
   B = Rf + 2.0*X + Rm*(sqrt(3.0)*cos(beta ) + sin(beta))* (cos(theta)*cos(psi)*sin(phi) + cos(phi)*sin(psi))+2.0*Rm*sin(PI/6.0 -beta)*(-cos(phi)*cos(psi) + cos(theta)*sin(phi)*sin(psi)) ;
   C =  4.0*pow(Lb,2) - 4.0*pow(Z - Rm*cos(PI/6.0-beta-psi)*sin(theta),2) ; 
   a = 4.0 ; 
   b = 2.0*A - 2.0*sqrt(3.0)*B ; 
   c = pow(A,2) + pow(B,2) - C ; 
   IOcheck(a, b, c, 5, Lv) ; 

// Lv contient la distance x entre le centre de la glissiere et la rotule 
// -> On convertit en longueur de vérin avec l'offset L0
   for (int i=0; i<6; i++) {
     Lv[i] = Lv[i] + L0 ;
   }  
}
// ============================================================================================

// =========================== MESURE LA POSITION ============================================= 
// Convertit le potentiel mesuré en longueur vérin
// Arguments : - potentiel
// Returns   : - longueur vérin en mm

float pot2pos(int pot){
  float posmes ; 
  posmes = 150.0*pot/1023.0  ;
  return posmes ; 
}// ============================================================================================

// =========================== IMPOSE LA COMMANDE MOTEUR ======================================= 
// Impose la tension souhaitée au moteur i
// Arguments : niveau souhaité dans [ -255 ; 255 ]
// Sortie    : rien

void Set_speed(int im, int com) {
  // Sens négatif
  if(com<0) {
    digitalWrite(pinI1I2[2*im  ],HIGH);
    digitalWrite(pinI1I2[2*im+1],LOW) ;
  }
  // Sens positif
  else {
    digitalWrite(pinI1I2[2*im  ],LOW);
    digitalWrite(pinI1I2[2*im+1],HIGH) ;
  } 
  // Niveau
  analogWrite(speedpin[im],abs(com));
}
// ============================================================================================

// =========================== BOUCLE INFINIE PRINCIPALE ======================================
void loop() {
  // Liste contenant les longueurs des verins actuelles
  float Lv[6];
  
  // Somme de toutes les erreurs en position des verins
  float errtot ;
  
  // Booleen pilotant l'arret de la boucle d'asservissement
  boolean nonstop; 
  
  // Temps passe a realiser l'asservissement
  unsigned long time_start ;
  unsigned long time_elapsed ;
  
  // Position mesuree sur un verin  
  float posmes ;

  // On lit ce qui arrive sur le port série et on décode si besoin (mise à jour des paramètres par ex)
  serial_input_read() ; 
  if (Ser_Inp_Str_End) { 
    serial_input_decode();
  } 

  // Si le Start a été écrit, on lance le travail
  if (start) {
    
    // On construit la trajectoire
    build_traj();

    // Boucle sur les n points de la trajectoire
    for (int ip=0; ip<n; ip++){
      
      // Calcul des longueurs vérin via la loi entrée 
      IOLaw(tabposx[ip], tabposy[ip], tabposz[ip],tabposth[ip], tabposph[ip], tabposps[ip], Lv) ; 

      // Impression de la consigne sur le port série (pour interface)
      Serial.print("* DEB-PAS ") ;
      Serial.print(" : ") ;
      Serial.print(tabposx[ip] ) ;
      Serial.print(" : ") ;
      Serial.print(tabposy[ip] ) ;
      Serial.print(" : ") ;
      Serial.print(tabposz[ip] ) ; 
      Serial.print(" : ") ;
      Serial.print(tabposth[ip]) ;
      Serial.print(" : ") ;
      Serial.print(tabposph[ip]) ;
      Serial.print(" : ") ;
      Serial.println(tabposps[ip]) ;

      // Récupération du temps de début de l'asservissement sur le point courant
      time_start = millis();
      
      // Boucle sur le moteurs pour initialisation
      for (int im=0; im<6; im++){
          
          // Initalisation de l'erreur cumulée (pour l'action intégrale
          integr[im]=0.0; 
          
          // Mesure des positions courantes 
          pospot = analogRead(pinpot[im]) ;
          posmes = pot2pos(pospot) ; 
          posmes = posmes + Lmin ;

          // Erreur actuelle 
          error0[im] = Lv[im] -  posmes ;
         
      }
      
      // Initialisation du booleen qui pourra stopper la boucle d'asservissement
      nonstop = true ; 

      // Boucle d'asservissement tant que l'un des critères d'arrêt n'est pas atteints
      while (nonstop){

          // Initialisation de l'erreur totale
          errtot = 0. ;
    
          // Boucle sur les 6 vérins 
          for (int im=0; im<6; im++){
            
            // Lecture de la position
            pospot = analogRead(pinpot[im]) ;
            posmes = pot2pos(pospot) ; 
            posmes = posmes + Lmin ;
        
            // Calcul du correcteur PID 
            error[im] = Lv[im] -  posmes ;
            integr[im] += (error[im]+error0[im])*Te/2.0 ;  
            deriv[im] = (error[im]-error0[im])/Te;

            // Sauvegarde de l'erreur pour le pas de temps suivant 
            error0[im] = error[im];
            
            // Saturation de l'intégrale   
            if(integr[im]>INTG_MAX)
              {integr[im]=INTG_MAX;}
            if(integr[im]<-INTG_MAX)
              {integr[im]=-INTG_MAX;}  
    
            // Valeur de la commande moteur
            com = KP[im]*error[im] + KI[im]*integr[im] + KD[im]*deriv[im]; 
      
            // Saturation de la commande
            if(com>255)
              {com=255;}
            if(com<-255)
              {com=-255;}
    
            // Ecrit sur le port série pour interface : Longueur mesuree, longueur actuelle, commande
            //Serial.print(" --> Actuator number : ") ;
            Serial.print(im+1) ; 
            Serial.print(" : ");
            //Serial.print(" - D-Length (mm)  : ") ;
            Serial.print(Lv[im]) ;  
            Serial.print(" : ");
            //Serial.print(" - M-Length (mm)  : ") ;
            Serial.print(posmes);
            Serial.print(" : ");
            //Serial.print(" - Command (-)  : ") ;
            Serial.println(com);    

            // On applique la commande moteur au vérin im
            Set_speed(im, com) ;

            // Calcule l'erreur globale sur tous les axes
            errtot = errtot + pow(error[im],2) ; 
          } 
  
          // Erreur globale pour la position souhaitée
          errtot = sqrt(errtot) ;
      
          // Temps déjà passé pour atteindre la position souhaitée
          time_elapsed = millis() - time_start ; 
                        
          // Evaluation des critères d'arrêt de la boucle d'asservissement
          if ( (errtot < TOL_ERR) || (time_elapsed > Tempo) ){
            nonstop = false;
            //Serial.print("* Erreur a la fin :") ;
            //Serial.print("\t") ; 
            //Serial.println(errtot);
            Serial.println("* FIN-PAS") ;
          }

      //  Fin de la boucle d'asservissement
      }
      
    //  Fin de la boucle sur les points de la trajectoire
    }
    
    // La trajectoire est terminée, on attendra un nouveau Start
    start = false ;

    // Les moteurs sont arrêtés 
    for (int im=0; im<6; im++){
      Set_speed(im,0);
    }

  // Fin du IF Start
  } 
  
}
// ============================================================================================
